@@ -1,5 +1,9 @@
 """
-The :mod:`tslearn.metrics` module gathers time series similarity metrics.
+The :mod:`tslearn.metrics` module delivers time-series specific metrics to be 
+used at the core of machine learning algorithms.
+
+**User guide:** See the :ref:`Dynamic Time Warping (DTW) <dtw>` section for 
+further details.
 """
 
 import warnings
@@ -8,10 +12,11 @@ import numpy
 from joblib import Parallel, delayed
 from numba import njit, prange
 from scipy.spatial.distance import pdist, cdist
-from sklearn.metrics.pairwise import euclidean_distances
+from sklearn.metrics.pairwise import euclidean_distances, pairwise_distances
 from sklearn.utils import check_random_state
 from tslearn.soft_dtw_fast import _soft_dtw, _soft_dtw_grad, \
     _jacobian_product_sq_euc
+from tslearn.cysax import cydist_sax
 
 from tslearn.utils import to_time_series, to_time_series_dataset, ts_size, \
     check_equal_size
@@ -19,13 +24,14 @@ from tslearn.utils import to_time_series, to_time_series_dataset, ts_size, \
 __author__ = 'Romain Tavenard romain.tavenard[at]univ-rennes2.fr'
 
 GLOBAL_CONSTRAINT_CODE = {None: 0, "": 0, "itakura": 1, "sakoe_chiba": 2}
-VARIABLE_LENGTH_METRICS = ["dtw", "gak", "softdtw"]
+TSLEARN_VALID_METRICS = ["dtw", "gak", "softdtw", "sax"]
+VARIABLE_LENGTH_METRICS = ["dtw", "gak", "softdtw", "sax"]
 
 
 @njit()
 def _local_squared_dist(x, y):
     dist = 0.
-    for di in prange(x.shape[0]):
+    for di in range(x.shape[0]):
         diff = (x[di] - y[di])
         dist += diff * diff
     return dist
@@ -57,8 +63,8 @@ def njit_accumulated_matrix(s1, s2, mask):
     cum_sum = numpy.full((l1 + 1, l2 + 1), numpy.inf)
     cum_sum[0, 0] = 0.
 
-    for i in prange(l1):
-        for j in prange(l2):
+    for i in range(l1):
+        for j in range(l2):
             if numpy.isfinite(mask[i, j]):
                 cum_sum[i + 1, j + 1] = _local_squared_dist(s1[i], s2[j])
                 cum_sum[i + 1, j + 1] += min(cum_sum[i, j + 1],
@@ -123,14 +129,15 @@ def dtw_path(s1, s2, global_constraint=None, sakoe_chiba_radius=None,
     similarity.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the alignment path:
+    i.e., if :math:`\pi` is the alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} (X_{i} - Y_{j})^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} (X_{i} - Y_{j})^2}
 
     It is not required that both time series share the same size, but they must
-    be the same dimension. DTW was originally presented in [1]_.
+    be the same dimension. DTW was originally presented in [1]_ and is
+    discussed in more details in our :ref:`dedicated user-guide page <dtw>`.
 
     Parameters
     ----------
@@ -187,6 +194,7 @@ def dtw_path(s1, s2, global_constraint=None, sakoe_chiba_radius=None,
     --------
     dtw : Get only the similarity score for DTW
     cdist_dtw : Cross similarity matrix between time series datasets
+    dtw_path_from_metric : Compute a DTW using a user-defined distance metric
 
     References
     ----------
@@ -207,22 +215,206 @@ def dtw_path(s1, s2, global_constraint=None, sakoe_chiba_radius=None,
     return path, numpy.sqrt(acc_cost_mat[-1, -1])
 
 
+@njit()
+def njit_accumulated_matrix_from_dist_matrix(dist_matrix, mask):
+    """Compute the accumulated cost matrix score between two time series using
+    a precomputed distance matrix.
+
+    Parameters
+    ----------
+    dist_matrix : array, shape = (sz1, sz2)
+        Array containing the pairwise distances.
+
+    mask : array, shape = (sz1, sz2)
+        Mask. Unconsidered cells must have infinite values.
+
+    Returns
+    -------
+    mat : array, shape = (sz1, sz2)
+        Accumulated cost matrix.
+
+    """
+    l1, l2 = dist_matrix.shape
+    cum_sum = numpy.full((l1 + 1, l2 + 1), numpy.inf)
+    cum_sum[0, 0] = 0.
+
+    for i in prange(l1):
+        for j in prange(l2):
+            if numpy.isfinite(mask[i, j]):
+                cum_sum[i + 1, j + 1] = dist_matrix[i, j]
+                cum_sum[i + 1, j + 1] += min(cum_sum[i, j + 1],
+                                             cum_sum[i + 1, j],
+                                             cum_sum[i, j])
+    return cum_sum[1:, 1:]
+
+
+def dtw_path_from_metric(s1, s2=None, metric="euclidean",
+                         global_constraint=None, sakoe_chiba_radius=None,
+                         itakura_max_slope=None, **kwds):
+    r"""Compute Dynamic Time Warping (DTW) similarity measure between
+    (possibly multidimensional) time series using a distance metric defined by
+    the user and return both the path and the similarity.
+
+    Similarity is computed as the cumulative cost along the aligned time
+    series.
+
+    It is not required that both time series share the same size, but they must
+    be the same dimension. DTW was originally presented in [1]_.
+
+    Valid values for metric are the same as for scikit-learn
+    `pairwise_distances`_ function i.e. a string (e.g. "euclidean",
+    "sqeuclidean", "hamming") or a function that is used to compute the
+    pairwise distances. See `scikit`_ and `scipy`_ documentations for more
+    information about the available metrics.
+
+    Parameters
+    ----------
+    s1 : array, shape = (sz1, d) if metric!="precomputed", (sz1, sz2) otherwise
+        A time series or an array of pairwise distances between samples.
+
+    s2 : array, shape = (sz2, d), optional (default: None)
+        A second time series, only allowed if metric != "precomputed".
+
+    metric : string or callable (default: "euclidean")
+        Function used to compute the pairwise distances between each points of
+        `s1` and `s2`.
+
+        If metric is "precomputed", `s1` is assumed to be a distance matrix.
+
+        If metric is an other string, it must be one of the options compatible
+        with sklearn.metrics.pairwise_distances.
+
+        Alternatively, if metric is a callable function, it is called on pairs
+        of rows of `s1` and `s2`. The callable should take two 1 dimensional
+        arrays as input and return a value indicating the distance between
+        them.
+
+    global_constraint : {"itakura", "sakoe_chiba"} or None (default: None)
+        Global constraint to restrict admissible paths for DTW.
+
+    sakoe_chiba_radius : int or None (default: None)
+        Radius to be used for Sakoe-Chiba band global constraint.
+        If None and `global_constraint` is set to "sakoe_chiba", a radius of
+        1 is used.
+        If both `sakoe_chiba_radius` and `itakura_max_slope` are set,
+        `global_constraint` is used to infer which constraint to use among the
+        two. In this case, if `global_constraint` corresponds to no global
+        constraint, a `RuntimeWarning` is raised and no global constraint is
+        used.
+
+    itakura_max_slope : float or None (default: None)
+        Maximum slope for the Itakura parallelogram constraint.
+        If None and `global_constraint` is set to "itakura", a maximum slope
+        of 2. is used.
+        If both `sakoe_chiba_radius` and `itakura_max_slope` are set,
+        `global_constraint` is used to infer which constraint to use among the
+        two. In this case, if `global_constraint` corresponds to no global
+        constraint, a `RuntimeWarning` is raised and no global constraint is
+        used.
+
+    **kwds
+        Additional arguments to pass to sklearn pairwise_distances to compute
+        the pairwise distances.
+
+    Returns
+    -------
+    list of integer pairs
+        Matching path represented as a list of index pairs. In each pair, the
+        first index corresponds to s1 and the second one corresponds to s2.
+
+    float
+        Similarity score (sum of metric along the wrapped time series).
+
+    Examples
+    --------
+    Lets create 2 numpy arrays to wrap:
+
+    >>> import numpy as np
+    >>> rng = np.random.RandomState(0)
+    >>> s1, s2 = rng.rand(5, 2), rng.rand(6, 2)
+
+    The wrapping can be done by passing a string indicating the metric to pass
+    to scikit-learn pairwise_distances:
+
+    >>> dtw_path_from_metric(s1, s2,
+    ...                      metric="sqeuclidean")  # doctest: +ELLIPSIS
+    ([(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5)], 1.117...)
+
+    Or by defining a custom distance function:
+
+    >>> sqeuclidean = lambda x, y: np.sum((x-y)**2)
+    >>> dtw_path_from_metric(s1, s2, metric=sqeuclidean)  # doctest: +ELLIPSIS
+    ([(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5)], 1.117...)
+
+    Or by using a precomputed distance matrix as input:
+
+    >>> from sklearn.metrics.pairwise import pairwise_distances
+    >>> dist_matrix = pairwise_distances(s1, s2, metric="sqeuclidean")
+    >>> dtw_path_from_metric(dist_matrix,
+    ...                      metric="precomputed")  # doctest: +ELLIPSIS
+    ([(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5)], 1.117...)
+
+    Notes
+    --------
+    By using a squared euclidean distance metric as shown above, the output
+    path is the same as the one obtained by using dtw_path but the similarity
+    score is the sum of squared distances instead of the euclidean distance.
+
+    See Also
+    --------
+    dtw_path : Get both the matching path and the similarity score for DTW
+
+    References
+    ----------
+    .. [1] H. Sakoe, S. Chiba, "Dynamic programming algorithm optimization for
+           spoken word recognition," IEEE Transactions on Acoustics, Speech and
+           Signal Processing, vol. 26(1), pp. 43--49, 1978.
+
+    .. _pairwise_distances: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html
+
+    .. _scikit: https://scikit-learn.org/stable/modules/metrics.html
+
+    .. _scipy: https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html
+
+    """  # noqa: E501
+    if metric == "precomputed":  # Pairwise distance given as input
+        sz1, sz2 = s1.shape
+        mask = compute_mask(
+            sz1, sz2, GLOBAL_CONSTRAINT_CODE[global_constraint],
+            sakoe_chiba_radius, itakura_max_slope
+        )
+        dist_mat = s1
+    else:
+        s1 = to_time_series(s1, remove_nans=True)
+        s2 = to_time_series(s2, remove_nans=True)
+        mask = compute_mask(
+            s1, s2, GLOBAL_CONSTRAINT_CODE[global_constraint],
+            sakoe_chiba_radius, itakura_max_slope
+        )
+        dist_mat = pairwise_distances(s1, s2, metric=metric, **kwds)
+
+    acc_cost_mat = njit_accumulated_matrix_from_dist_matrix(dist_mat, mask)
+    path = _return_path(acc_cost_mat)
+    return path, acc_cost_mat[-1, -1]
+
+
 def dtw(s1, s2, global_constraint=None, sakoe_chiba_radius=None,
         itakura_max_slope=None):
     r"""Compute Dynamic Time Warping (DTW) similarity measure between
     (possibly multidimensional) time series and return it.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the optimal alignment path:
+    i.e., if :math:`\pi` is the optimal alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} \|X_{i} - Y_{j}\|^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} \|X_{i} - Y_{j}\|^2}
 
     Note that this formula is still valid for the multivariate case.
 
     It is not required that both time series share the same size, but they must
-    be the same dimension. DTW was originally presented in [1]_.
+    be the same dimension. DTW was originally presented in [1]_ and is
+    discussed in more details in our :ref:`dedicated user-guide page <dtw>`.
 
     Parameters
     ----------
@@ -377,17 +569,19 @@ def dtw_limited_warping_length(s1, s2, max_length):
     the resulting path length and return the similarity cost.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the optimal alignment path:
+    i.e., if :math:`\pi` is the optimal alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} \|X_{i} - Y_{j}\|^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} \|X_{i} - Y_{j}\|^2}
 
     Note that this formula is still valid for the multivariate case.
 
     It is not required that both time series share the same size, but they must
     be the same dimension. DTW was originally presented in [1]_.
     This constrained-length variant was introduced in [2]_.
+    Both bariants are
+    discussed in more details in our :ref:`dedicated user-guide page <dtw>`
 
     Parameters
     ----------
@@ -478,17 +672,19 @@ def dtw_path_limited_warping_length(s1, s2, max_length):
     cost.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the optimal alignment path:
+    i.e., if :math:`\pi` is the optimal alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} \|X_{i} - Y_{j}\|^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} \|X_{i} - Y_{j}\|^2}
 
     Note that this formula is still valid for the multivariate case.
 
     It is not required that both time series share the same size, but they must
     be the same dimension. DTW was originally presented in [1]_.
     This constrained-length variant was introduced in [2]_.
+    Both variants are
+    discussed in more details in our :ref:`dedicated user-guide page <dtw>`
 
     Parameters
     ----------
@@ -571,8 +767,8 @@ def _subsequence_cost_matrix(subseq, longseq):
     cum_sum = numpy.full((l1 + 1, l2 + 1), numpy.inf)
     cum_sum[0, :] = 0.
 
-    for i in prange(l1):
-        for j in prange(l2):
+    for i in range(l1):
+        for j in range(l2):
             cum_sum[i + 1, j + 1] = _local_squared_dist(subseq[i], longseq[j])
             cum_sum[i + 1, j + 1] += min(cum_sum[i, j + 1],
                                          cum_sum[i + 1, j],
@@ -625,13 +821,14 @@ def _subsequence_path(acc_cost_mat, idx_path_end):
 
 
 def subsequence_path(acc_cost_mat, idx_path_end):
-    """Compute the optimal path through a accumulated cost matrix given the
+    r"""Compute the optimal path through a accumulated cost matrix given the
     endpoint of the sequence.
 
     Parameters
     ----------
     acc_cost_mat: array, shape = (sz1, sz2)
-        The accumulated cost matrix comparing subsequence from a longer sequence
+        The accumulated cost matrix comparing subsequence from a longer
+        sequence.
     idx_path_end: int
         The end position of the matched subsequence in the longer sequence.
 
@@ -669,16 +866,16 @@ def dtw_subsequence_path(subseq, longseq):
     return both the path and the similarity.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the alignment path:
+    i.e., if :math:`\pi` is the alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} (X_{i} - Y_{j})^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} \|X_{i} - Y_{j}\|^2}
 
     Compared to traditional DTW, here, border constraints on admissible paths
-    :math:`P` are relaxed such that :math:`P_0 = (0, ?)` and
-    :math:`P_L = (N-1, ?)` where :math:`L` is the length of the considered path
-    and :math:`N` is the length of the subsequence time series.
+    :math:`\pi` are relaxed such that :math:`\pi_0 = (0, ?)` and
+    :math:`\pi_L = (N-1, ?)` where :math:`L` is the length of the considered
+    path and :math:`N` is the length of the subsequence time series.
 
     It is not required that both time series share the same size, but they must
     be the same dimension. This implementation finds the best matching starting
@@ -884,10 +1081,10 @@ def compute_mask(s1, s2, global_constraint=0,
     Parameters
     ----------
     s1 : array
-        A time series.
+        A time series or integer.
 
     s2: array
-        Another time series.
+        Another time series or integer.
 
     global_constraint : {0, 1, 2} (default: 0)
         Global constraint to restrict admissible paths for DTW:
@@ -920,8 +1117,12 @@ def compute_mask(s1, s2, global_constraint=0,
     mask : array
         Constraint region.
     """
-    sz1 = s1.shape[0]
-    sz2 = s2.shape[0]
+    # The output mask will be of shape (sz1, sz2)
+    if isinstance(s1, int) and isinstance(s2, int):
+        sz1, sz2 = s1, s2
+    else:
+        sz1 = s1.shape[0]
+        sz2 = s2.shape[0]
     if (global_constraint == 0 and sakoe_chiba_radius is not None
             and itakura_max_slope is not None):
         raise RuntimeWarning("global_constraint is not set for DTW, but both "
@@ -943,19 +1144,101 @@ def compute_mask(s1, s2, global_constraint=0,
     return mask
 
 
+def _cdist_generic(dist_fun, dataset1, dataset2, n_jobs, verbose,
+                   compute_diagonal=True, dtype=numpy.float, *args, **kwargs):
+    """Compute cross-similarity matrix with joblib parallelization for a given
+    similarity function.
+
+    Parameters
+    ----------
+    dist_fun : function
+        Similarity function to be used
+
+    dataset1 : array-like
+        A dataset of time series
+
+    dataset2 : array-like (default: None)
+        Another dataset of time series. If `None`, self-similarity of
+        `dataset1` is returned.
+
+    n_jobs : int or None, optional (default=None)
+        The number of jobs to run in parallel.
+        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
+        ``-1`` means using all processors. See scikit-learns'
+        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`__
+        for more details.
+
+    verbose : int, optional (default=0)
+        The verbosity level: if non zero, progress messages are printed.
+        Above 50, the output is sent to stdout.
+        The frequency of the messages increases with the verbosity level.
+        If it more than 10, all iterations are reported.
+        `Glossary <https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation>`__
+        for more details.
+
+    compute_diagonal : bool (default: True)
+        Whether diagonal terms should be computed or assumed to be 0 in the
+        self-similarity case. Used only if `dataset2` is `None`.
+
+    *args and **kwargs :
+        Optional additional parameters to be passed to the similarity function.
+
+
+    Returns
+    -------
+    cdist : numpy.ndarray
+        Cross-similarity matrix
+    """  # noqa: E501
+    dataset1 = to_time_series_dataset(dataset1, dtype=dtype)
+
+    if dataset2 is None:
+        # Inspired from code by @GillesVandewiele:
+        # https://github.com/rtavenar/tslearn/pull/128#discussion_r314978479
+        matrix = numpy.zeros((len(dataset1), len(dataset1)))
+        indices = numpy.triu_indices(len(dataset1),
+                                     k=0 if compute_diagonal else 1,
+                                     m=len(dataset1))
+        matrix[indices] = Parallel(n_jobs=n_jobs,
+                                   prefer="threads",
+                                   verbose=verbose)(
+            delayed(dist_fun)(
+                dataset1[i], dataset1[j],
+                *args, **kwargs
+            )
+            for i in range(len(dataset1))
+            for j in range(i if compute_diagonal else i + 1,
+                           len(dataset1))
+        )
+        indices = numpy.tril_indices(len(dataset1), k=-1, m=len(dataset1))
+        matrix[indices] = matrix.T[indices]
+        return matrix
+    else:
+        dataset2 = to_time_series_dataset(dataset2, dtype=dtype)
+        matrix = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
+            delayed(dist_fun)(
+                dataset1[i], dataset2[j],
+                *args, **kwargs
+            )
+            for i in range(len(dataset1)) for j in range(len(dataset2))
+        )
+        return numpy.array(matrix).reshape((len(dataset1), -1))
+
+
 def cdist_dtw(dataset1, dataset2=None, global_constraint=None,
-              sakoe_chiba_radius=None, itakura_max_slope=None, n_jobs=None, verbose=0):
+              sakoe_chiba_radius=None, itakura_max_slope=None, n_jobs=None,
+              verbose=0):
     r"""Compute cross-similarity matrix using Dynamic Time Warping (DTW)
     similarity measure.
 
     DTW is computed as the Euclidean distance between aligned time series,
-    i.e., if :math:`P` is the alignment path:
+    i.e., if :math:`\pi` is the alignment path:
 
     .. math::
 
-        DTW(X, Y) = \sqrt{\sum_{(i, j) \in P} (X_{i} - Y_{j})^2}
+        DTW(X, Y) = \sqrt{\sum_{(i, j) \in \pi} (X_{i} - Y_{j})^2}
 
-    DTW was originally presented in [1]_.
+    DTW was originally presented in [1]_ and is
+    discussed in more details in our :ref:`dedicated user-guide page <dtw>`.
 
     Parameters
     ----------
@@ -993,13 +1276,13 @@ def cdist_dtw(dataset1, dataset2=None, global_constraint=None,
         The number of jobs to run in parallel.
         ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
         ``-1`` means using all processors. See scikit-learns'
-        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`_
+        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`__
         for more details.
 
     verbose : int, optional (default=0)
-        The verbosity level: if non zero, progress messages are printed. 
-        Above 50, the output is sent to stdout. 
-        The frequency of the messages increases with the verbosity level. 
+        The verbosity level: if non zero, progress messages are printed.
+        Above 50, the output is sent to stdout.
+        The frequency of the messages increases with the verbosity level.
         If it more than 10, all iterations are reported.
         `Glossary <https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation>`__
         for more details.
@@ -1027,35 +1310,13 @@ def cdist_dtw(dataset1, dataset2=None, global_constraint=None,
     .. [1] H. Sakoe, S. Chiba, "Dynamic programming algorithm optimization for
            spoken word recognition," IEEE Transactions on Acoustics, Speech and
            Signal Processing, vol. 26(1), pp. 43--49, 1978.
-
-    """ # noqa: E501
-    dataset1 = to_time_series_dataset(dataset1)
-
-    if dataset2 is None:
-        # Inspired from code by @GillesVandewiele:
-        # https://github.com/rtavenar/tslearn/pull/128#discussion_r314978479
-        matrix = numpy.zeros((len(dataset1), len(dataset1)))
-        indices = numpy.triu_indices(len(dataset1), k=1, m=len(dataset1))
-        matrix[indices] = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
-            delayed(dtw)(
-                dataset1[i], dataset1[j],
-                global_constraint=global_constraint,
-                sakoe_chiba_radius=sakoe_chiba_radius,
-                itakura_max_slope=itakura_max_slope)
-            for i in range(len(dataset1)) for j in range(i + 1, len(dataset1))
-        )
-        return matrix + matrix.T
-    else:
-        dataset2 = to_time_series_dataset(dataset2)
-        matrix = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
-            delayed(dtw)(
-                dataset1[i], dataset2[j],
-                global_constraint=global_constraint,
-                sakoe_chiba_radius=sakoe_chiba_radius,
-                itakura_max_slope=itakura_max_slope)
-            for i in range(len(dataset1)) for j in range(len(dataset2))
-        )
-        return numpy.array(matrix).reshape((len(dataset1), -1))
+    """  # noqa: E501
+    return _cdist_generic(dist_fun=dtw, dataset1=dataset1, dataset2=dataset2,
+                          n_jobs=n_jobs, verbose=verbose,
+                          compute_diagonal=False,
+                          global_constraint=global_constraint,
+                          sakoe_chiba_radius=sakoe_chiba_radius,
+                          itakura_max_slope=itakura_max_slope)
 
 
 @njit(nogil=True)
@@ -1066,8 +1327,8 @@ def njit_gak(s1, s2, gram):
     cum_sum = numpy.zeros((l1 + 1, l2 + 1))
     cum_sum[0, 0] = 1.
 
-    for i in prange(l1):
-        for j in prange(l2):
+    for i in range(l1):
+        for j in range(l2):
             cum_sum[i + 1, j + 1] = (cum_sum[i, j + 1] +
                                      cum_sum[i + 1, j] +
                                      cum_sum[i, j]) * gram[i, j]
@@ -1193,12 +1454,12 @@ def cdist_gak(dataset1, dataset2=None, sigma=1., n_jobs=None, verbose=0):
         The number of jobs to run in parallel.
         ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
         ``-1`` means using all processors. See scikit-learns'
-        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`_
+        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`__
         for more details.
     verbose : int, optional (default=0)
-        The verbosity level: if non zero, progress messages are printed. 
-        Above 50, the output is sent to stdout. 
-        The frequency of the messages increases with the verbosity level. 
+        The verbosity level: if non zero, progress messages are printed.
+        Above 50, the output is sent to stdout.
+        The frequency of the messages increases with the verbosity level.
         If it more than 10, all iterations are reported.
         `Glossary <https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation>`__
         for more details.
@@ -1226,40 +1487,35 @@ def cdist_gak(dataset1, dataset2=None, sigma=1., n_jobs=None, verbose=0):
     References
     ----------
     .. [1] M. Cuturi, "Fast global alignment kernels," ICML 2011.
-    """ # noqa: E501
+    """  # noqa: E501
+    unnormalized_matrix = _cdist_generic(dist_fun=unnormalized_gak,
+                                         dataset1=dataset1,
+                                         dataset2=dataset2,
+                                         n_jobs=n_jobs,
+                                         verbose=verbose,
+                                         sigma=sigma,
+                                         compute_diagonal=True)
     dataset1 = to_time_series_dataset(dataset1)
-
     if dataset2 is None:
-        # Inspired from code by @GillesVandewiele:
-        # https://github.com/rtavenar/tslearn/pull/128#discussion_r314978479
-        matrix = numpy.zeros((len(dataset1), len(dataset1)))
-        indices = numpy.triu_indices(len(dataset1), k=0, m=len(dataset1))
-        matrix[indices] = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
-            delayed(unnormalized_gak)(dataset1[i], dataset1[j], sigma=sigma)
-            for i in range(len(dataset1)) for j in range(i, len(dataset1))
-        )
-        indices = numpy.tril_indices(len(dataset1), k=-1, m=len(dataset1))
-        matrix[indices] = matrix.T[indices]
-        diagonal = numpy.diag(numpy.sqrt(1. / numpy.diag(matrix)))
+        diagonal = numpy.diag(numpy.sqrt(1. / numpy.diag(unnormalized_matrix)))
         diagonal_left = diagonal_right = diagonal
     else:
         dataset2 = to_time_series_dataset(dataset2)
-        matrix = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
-            delayed(unnormalized_gak)(dataset1[i], dataset2[j], sigma=sigma)
-            for i in range(len(dataset1)) for j in range(len(dataset2))
-        )
-        matrix = numpy.array(matrix).reshape((len(dataset1), -1))
-        diagonal_left = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
+        diagonal_left = Parallel(n_jobs=n_jobs,
+                                 prefer="threads",
+                                 verbose=verbose)(
             delayed(unnormalized_gak)(dataset1[i], dataset1[i], sigma=sigma)
             for i in range(len(dataset1))
         )
-        diagonal_right = Parallel(n_jobs=n_jobs, prefer="threads", verbose=verbose)(
+        diagonal_right = Parallel(n_jobs=n_jobs,
+                                  prefer="threads",
+                                  verbose=verbose)(
             delayed(unnormalized_gak)(dataset2[j], dataset2[j], sigma=sigma)
             for j in range(len(dataset2))
         )
         diagonal_left = numpy.diag(1. / numpy.sqrt(diagonal_left))
         diagonal_right = numpy.diag(1. / numpy.sqrt(diagonal_right))
-    return (diagonal_left.dot(matrix)).dot(diagonal_right)
+    return (diagonal_left.dot(unnormalized_matrix)).dot(diagonal_right)
 
 
 def sigma_gak(dataset, n_samples=100, random_state=None):
@@ -1280,7 +1536,7 @@ def sigma_gak(dataset, n_samples=100, random_state=None):
     Returns
     -------
     float
-        Suggested bandwidth (:math:`\\sigma`) for the Global Alignment kernel
+        Suggested bandwidth (:math:`\sigma`) for the Global Alignment kernel
 
     Examples
     --------
@@ -1334,7 +1590,7 @@ def gamma_soft_dtw(dataset, n_samples=100, random_state=None):
     Returns
     -------
     float
-        Suggested :math:`\\gamma` parameter for the Soft-DTW
+        Suggested :math:`\gamma` parameter for the Soft-DTW
 
     Examples
     --------
@@ -1357,6 +1613,60 @@ def gamma_soft_dtw(dataset, n_samples=100, random_state=None):
                           random_state=random_state) ** 2
 
 
+def cdist_sax(dataset1, breakpoints_avg, size_fitted, dataset2=None,
+              n_jobs=None, verbose=0):
+    r"""Calculates a matrix of distances (MINDIST) on SAX-transformed data,
+    as presented in [1]_. It is important to note that this function
+    expects the timeseries in dataset1 and dataset2 to be normalized
+    to each have zero mean and unit variance.
+
+    Parameters
+    ----------
+    dataset1 : array-like
+        A dataset of time series
+
+    breakpoints_avg : array-like
+        The breakpoints used to assign the alphabet symbols.
+
+    size_fitted: int
+        The original timesteps in the timeseries, before
+        discretizing through SAX.
+
+    dataset2 : array-like (default: None)
+        Another dataset of time series. If `None`, self-similarity of
+        `dataset1` is returned.
+
+    n_jobs : int or None, optional (default=None)
+        The number of jobs to run in parallel.
+        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
+        ``-1`` means using all processors. See scikit-learns'
+        `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`__
+        for more details.
+
+    verbose : int, optional (default=0)
+        The verbosity level: if non zero, progress messages are printed.
+        Above 50, the output is sent to stdout.
+        The frequency of the messages increases with the verbosity level.
+        If it more than 10, all iterations are reported.
+        `Glossary <https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation>`__
+        for more details.
+
+    Returns
+    -------
+    cdist : numpy.ndarray
+        Cross-similarity matrix
+
+    References
+    ----------
+    .. [1] Lin, Jessica, et al. "Experiencing SAX: a novel symbolic
+           representation of time series." Data Mining and knowledge
+           discovery 15.2 (2007): 107-144.
+
+    """  # noqa: E501
+    return _cdist_generic(cydist_sax, dataset1, dataset2, n_jobs, verbose,
+                          False, int, breakpoints_avg, size_fitted)
+
+
 def lb_keogh(ts_query, ts_candidate=None, radius=1, envelope_candidate=None):
     r"""Compute LB_Keogh.
 
@@ -1506,7 +1816,23 @@ def lb_envelope(ts, radius=1):
 def soft_dtw(ts1, ts2, gamma=1.):
     r"""Compute Soft-DTW metric between two time series.
 
-    Soft-DTW was originally presented in [1]_.
+    Soft-DTW was originally presented in [1]_ and is
+    discussed in more details in our
+    :ref:`user-guide page on DTW and its variants<dtw>`.
+
+    Soft-DTW is computed as:
+
+    .. math::
+
+        \text{soft-DTW}_{\gamma}(X, Y) =
+            \min_{\pi}{}^\gamma \sum_{(i, j) \in \pi} \|X_i, Y_j\|^2
+
+    where :math:`\min^\gamma` is the soft-min operator of parameter
+    :math:`\gamma`.
+
+    In the limit case :math:`\gamma = 0`, :math:`\min^\gamma` reduces to a
+    hard-min operator and soft-DTW is defined as the square of the DTW
+    similarity measure.
 
     Parameters
     ----------
@@ -1543,7 +1869,7 @@ def soft_dtw(ts1, ts2, gamma=1.):
        Time-Series," ICML 2017.
     """
     if gamma == 0.:
-        return dtw(ts1, ts2)
+        return dtw(ts1, ts2) ** 2
     return SoftDTW(SquaredEuclidean(ts1[:ts_size(ts1)], ts2[:ts_size(ts2)]),
                    gamma=gamma).compute()
 
@@ -1551,7 +1877,23 @@ def soft_dtw(ts1, ts2, gamma=1.):
 def cdist_soft_dtw(dataset1, dataset2=None, gamma=1.):
     r"""Compute cross-similarity matrix using Soft-DTW metric.
 
-    Soft-DTW was originally presented in [1]_.
+    Soft-DTW was originally presented in [1]_ and is
+    discussed in more details in our
+    :ref:`user-guide page on DTW and its variants<dtw>`.
+
+    Soft-DTW is computed as:
+
+    .. math::
+
+        \text{soft-DTW}_{\gamma}(X, Y) =
+            \min_{\pi}{}^\gamma \sum_{(i, j) \in \pi} \|X_i, Y_j\|^2
+
+    where :math:`\min^\gamma` is the soft-min operator of parameter
+    :math:`\gamma`.
+
+    In the limit case :math:`\gamma = 0`, :math:`\min^\gamma` reduces to a
+    hard-min operator and soft-DTW is defined as the square of the DTW
+    similarity measure.
 
     Parameters
     ----------
@@ -1617,21 +1959,47 @@ def cdist_soft_dtw(dataset1, dataset2=None, gamma=1.):
 
 
 def cdist_soft_dtw_normalized(dataset1, dataset2=None, gamma=1.):
-    """Compute cross-similarity matrix using a normalized version of the
+    r"""Compute cross-similarity matrix using a normalized version of the
     Soft-DTW metric.
 
-    Soft-DTW was originally presented in [1]_.
+    Soft-DTW was originally presented in [1]_ and is
+    discussed in more details in our
+    :ref:`user-guide page on DTW and its variants<dtw>`.
+
+    Soft-DTW is computed as:
+
+    .. math::
+
+        \text{soft-DTW}_{\gamma}(X, Y) =
+            \min_{\pi}{}^\gamma \sum_{(i, j) \in \pi} \|X_i, Y_j\|^2
+
+    where :math:`\min^\gamma` is the soft-min operator of parameter
+    :math:`\gamma`.
+
+    In the limit case :math:`\gamma = 0`, :math:`\min^\gamma` reduces to a
+    hard-min operator and soft-DTW is defined as the square of the DTW
+    similarity measure.
+
     This normalized version is defined as:
-    `sdtw_(x,y) := sdtw(x,y) - 1/2(sdtw(x,x)+sdtw(y,y))`
+
+    .. math::
+
+        \text{norm-soft-DTW}_{\gamma}(X, Y) =
+            \text{soft-DTW}_{\gamma}(X, Y) -
+            \frac{1}{2} \left(\text{soft-DTW}_{\gamma}(X, X) +
+                \text{soft-DTW}_{\gamma}(Y, Y)\right)
+
     and ensures that all returned values are positive and that
-    `sdtw_(x,x) == 0`.
+    :math:`\text{norm-soft-DTW}_{\gamma}(X, X) = 0`.
 
     Parameters
     ----------
     dataset1
         A dataset of time series
+
     dataset2
         Another dataset of time series
+
     gamma : float (default 1.)
         Gamma paraneter for Soft-DTW
 
@@ -1663,7 +2031,7 @@ def cdist_soft_dtw_normalized(dataset1, dataset2=None, gamma=1.):
     return dists
 
 
-class SoftDTW(object):
+class SoftDTW:
     def __init__(self, D, gamma=1.):
         """
         Parameters
@@ -1736,7 +2104,7 @@ class SoftDTW(object):
         return E[1:-1, 1:-1]
 
 
-class SquaredEuclidean(object):
+class SquaredEuclidean:
     def __init__(self, X, Y):
         """
         Parameters
